/*
 * libopen.c - loader for library
 *
 * Copyright (C) 2018 Thorsten Otto
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */
#include "slb/iconv.h"
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <time.h>
#include <stdint.h>
#include <errno.h>
#include <mint/cookie.h>
#include <mint/mintbind.h>
#include <math.h>
#include <setjmp.h>
#include "libstrct.h"
#include "slbload.h"

typedef struct
{
	long key;
	long value;
} COOKIE;


static COOKIE *get_cookie_jar(void)
{
	return (COOKIE *) Setexc(360, (void (*)()) -1L);
}


static int get_cookie(long key, long *value)
{
	COOKIE *cookiejar = get_cookie_jar();

	*value = 0;
	if (cookiejar)
	{
		while (cookiejar->key)
		{
			if (cookiejar->key == key)
			{
				*value = cookiejar->value;
				return 1;
			}
			cookiejar++;
		}
	}
	return 0;
}


#if defined(__MSHORT__) || defined(__PUREC__) || defined(__AHCC__)

static void *_CDECL iconv_gcc_memset(void *s, iconv_int_t val, size_t len)
{
	return memset(s, (int)val, len);
}

static void *_CDECL iconv_gcc_memcpy(void *d, const void *s, size_t len)
{
	return memcpy(d, s, len);
}

static iconv_int_t _CDECL iconv_gcc_memcmp(const void *s, const void *d, size_t len)
{
	return memcmp(s, d, len);
}

static size_t _CDECL iconv_gcc_strlen(const char *s)
{
	return strlen(s);
}

static char *_CDECL iconv_gcc_strcpy(char *d, const char *s)
{
	return strcpy(d, s);
}

static iconv_int_t _CDECL iconv_gcc_strcmp(const char *d, const char *s)
{
	return strcmp(d, s);
}

static char *_CDECL iconv_gcc_strchr(const char *d, iconv_int_t c)
{
	return strchr(d, (int)c);
}

static void *_CDECL iconv_gcc_malloc(size_t s)
{
	return malloc(s);
}

static void _CDECL iconv_gcc_free(void *s)
{
	free(s);
}

static iconv_int_t _CDECL iconv_gcc_get_errno(void)
{
#if defined(__GNUC__) || defined(_GNU_SOURCE)
	/* we are just here because of -mshort */
	return errno;
#elif defined(__PUREC__) || defined(__AHCC__) || defined(LATTICE)
	/* must translate error numbers from Pure-C library to what mintlib expects */
	switch (errno)
	{
		case EPERM: return 38;
		case ENOENT: return 33;
		case ESRCH: return 20;
		case EINTR: return 128;
		case EIO: return 90;
		case ENXIO: return 46;
		case EBADF: return 37;
#if defined(__PUREC__) || defined(__AHCC__)
		case EILLSPE: return 25; /* -> EINVAL */
		case EINVMEM: return 40; /* -> EFAULT */
#endif
		case EFAULT: return 40;
		case ENOTBLK: return 23;
		case EBUSY: return 2;
		case ENOMEM: return 39;
		case EACCES: return 36;
		case EEXIST: return 85;
#if defined(__PUREC__) || defined(__AHCC__)
		case EPLFMT: return 66; /* -> ENOEXEC */
#endif
#if defined(LATTICE)
		case ENOEXEC: return 66;
		case EAGAIN: return 326;
		case EXDEV: return 48;
		case ECHILD: return 48;
#endif
		case ENODEV: return 15;
		case ENOTDIR: return 34;
		case EINVAL: return 25;
		case ENFILE: return 50;
		case EMFILE: return 35;
		case ENOTTY: return 87;
		case ETXTBSY: return 70;
		case EFBIG: return 71;
		case ENOSPC: return 91;
		case ESPIPE: return 6;
		case EROFS: return 13;
		case EMLINK: return 82;
		case EPIPE: return 81;
		case EDOM: return 89;
		case ERANGE: return 88;
#if defined(__PUREC__) || defined(__AHCC__)
		case ENMFILE: return 49;
#endif
#if defined(LATTICE)
		case EDEADLK: return 22;
#endif
		case EILSEQ: return 27;
	}
	return 1;
#else
	return errno;
#endif
}

static void _CDECL iconv_gcc_set_errno(iconv_int_t e)
{
#if defined(__GNUC__) || defined(_GNU_SOURCE)
	/* we are just here because of -mshort */
	errno = (int)e;
#elif defined(__PUREC__) || defined(__AHCC__) || defined(LATTICE)
	/* must translate error numbers from mintlib to what Pure-C library expects */
	/* luckily, iconv only sets a few */
	switch ((int) e)
	{
		case 125: errno = E2BIG; break; /* E2BIG */
		case 27: errno = EILSEQ; break; /* EILSEQ */
		default: errno = (int)e; break;
	}
#else
	errno = (int)e;
#endif
}

static char *_CDECL iconv_gcc_getenv(const char *name)
{
	return getenv(name);
}

static iconv_int_t _CDECL (*qcompfunc)(const void *, const void *);

static int qcompar(const void *a, const void *b)
{
	return (int)qcompfunc(a, b);
}

static void _CDECL iconv_gcc_qsort(void *base, size_t nmemb, size_t size, iconv_int_t _CDECL (*compar)(const void *, const void *))
{
	qcompfunc = compar;
	qsort(base, nmemb, size, qcompar);
}

static size_t _CDECL iconv_gcc_wcrtomb(char *s, iconv_wchar_t wc, mbstate_t *st)
{
	(void) st;
	if (s == NULL)
		return 1;
	if ((uint32_t)wc < 0x80)
	{
		*s = wc;
		return 1;
	} else if ((uint32_t)wc < 0x800)
	{
		*s++ = 0xc0 | (wc>>6);
		*s = 0x80 | (wc&0x3f);
		return 2;
	} else if ((uint32_t)wc < 0xd800 || (unsigned)wc-0xe000 < 0x2000)
	{
		*s++ = 0xe0 | (wc>>12);
		*s++ = 0x80 | ((wc>>6)&0x3f);
		*s = 0x80 | (wc&0x3f);
		return 3;
	} else if ((uint32_t)wc-0x10000UL < 0x100000UL)
	{
		*s++ = 0xf0 | ((uint32_t)wc>>18);
		*s++ = 0x80 | ((wc>>12)&0x3f);
		*s++ = 0x80 | ((wc>>6)&0x3f);
		*s = 0x80 | (wc&0x3f);
		return 4;
	}
	errno = EILSEQ;
	return -1;
}

#define SA 0xc2u
#define SB 0xf4u
#define OOB(c,b) (((((b)>>3)-0x10)|(((b)>>3)+((int32_t)(c)>>26))) & ~7)
#define R(a,b) ((uint32_t)(a==0x80 ? 0x40 - b : -a) << 23)

#define C(x) ( x<2 ? (uint32_t)-1 : ( R(0x80,0xc0) | x ) )
#define D(x) C((x+16))
#define E(x) ( ( x==0 ? R(0xa0,0xc0) : \
                 x==0xd ? R(0x80,0xa0) : \
                 R(0x80,0xc0) ) \
             | ( R(0x80,0xc0) >> 6 ) \
             | x )
#define F(x) ( ( x>=5 ? 0 : \
                 x==0 ? R(0x90,0xc0) : \
                 x==4 ? R(0x80,0xa0) : \
                 R(0x80,0xc0) ) \
             | ( R(0x80,0xc0) >> 6 ) \
             | ( R(0x80,0xc0) >> 12 ) \
             | x )

static const uint32_t bittab[] = {
	              C(0x2),C(0x3),C(0x4),C(0x5),C(0x6),C(0x7),
	C(0x8),C(0x9),C(0xa),C(0xb),C(0xc),C(0xd),C(0xe),C(0xf),
	D(0x0),D(0x1),D(0x2),D(0x3),D(0x4),D(0x5),D(0x6),D(0x7),
	D(0x8),D(0x9),D(0xa),D(0xb),D(0xc),D(0xd),D(0xe),D(0xf),
	E(0x0),E(0x1),E(0x2),E(0x3),E(0x4),E(0x5),E(0x6),E(0x7),
	E(0x8),E(0x9),E(0xa),E(0xb),E(0xc),E(0xd),E(0xe),E(0xf),
	F(0x0),F(0x1),F(0x2),F(0x3),F(0x4)
};


static size_t _CDECL iconv_gcc_mbrtowc(wchar_t *wc, const char *src, size_t n, mbstate_t *st)
{
	static mbstate_t internal_state;
	uint32_t c;
	wchar_t dummy_wc;
	const unsigned char *s = (const unsigned char *)src;
	const unsigned int N = (unsigned int)n;

	if (st == NULL)
		st = &internal_state;
	c = st->__value.__wch;
	
	if (s == NULL)
	{
		if (c != 0)
			goto ilseq;
		return 0;
	}
	if (wc == NULL)
		wc = &dummy_wc;

	if (n == 0)
		return -2;
	if (c == 0)
	{
		if (*s < 0x80)
			return (*wc = *s) != 0;
		if (*s-SA > SB-SA)
			goto ilseq;
		c = bittab[*s++-SA];
		n--;
	}

	if (n != 0)
	{
		if (OOB(c, *s))
			goto ilseq;
loop:
		c = (c << 6) | (*s++ - 0x80);
		n--;
		if (!(c & (1UL << 31)))
		{
			st->__value.__wch = 0;
			*wc = c;
			return N - n;
		}
		if (n != 0)
		{
			if (*s-0x80u >= 0x40)
				goto ilseq;
			goto loop;
		}
	}

	st->__value.__wch = c;
	return -2;
ilseq:
	st->__value.__wch = 0;
	errno = EILSEQ;
	return -1;
}

static iconv_int_t __CDECL iconv_gcc_mbsinit(const mbstate_t *st)
{
	return st == NULL || *(const iconv_uint_t *)st == 0;
}

#define S(x) iconvslb_funcs.p_##x = iconv_gcc_##x

#else

static int _CDECL get_errno(void)
{
	return errno;
}

static void _CDECL set_errno(int e)
{
	errno = e;
}

#define S(x) iconvslb_funcs.p_##x = x

#endif


long slb_iconv_open(const char *slbpath)
{
	long ret;
	long flags;
	long cpu;
	
	SLB *iconvlib = slb_iconv_get();
	
	if (!iconvlib)
		return -EBADF;
	if (iconvlib->handle)
		return 0;

	iconvslb_funcs.struct_size = sizeof(iconvslb_funcs);
	iconvslb_funcs.int_size = sizeof(iconv_int_t),
	iconvslb_funcs.iconv_vernum = _LIBICONV_VERSION;

	S(memset);
	S(memcpy);
	S(memcmp);

	S(strlen);
	S(strcpy);
	S(strcmp);
	S(strchr);

	S(malloc);
	S(free);

	S(get_errno);
	S(set_errno);
	S(getenv);
	S(qsort);

	S(wcrtomb);
	S(mbrtowc);
	S(mbsinit);
	
#undef S
	
	ret = slb_load(SHAREDLIB, slbpath, _LIBICONV_VERSION, &iconvlib->handle, &iconvlib->exec);
	if (ret < 0)
	{
		slb_iconv_close();
		return ret;
	}

	/*
	 * check compile flags; that function should be as simple as to just return a constant
	 * and we can hopefully call it even on mismatched configurations
	 */
	flags = libiconv_slb_control(0, 0);
	get_cookie(C__CPU, &cpu);
	if (cpu >= 20)
	{
		/* should be able to use a 000 library, anyways */
	} else
	{
		if (flags & (1L << 16))
		{
			/* cpu is not 020+, but library was compiled for it */
			slb_iconv_close();
			return -EINVAL;
		}
	}
#if defined(__mcoldfire__)
	/* if cpu is cf, but library was not compiled for it... */
	if (!(flags & (1L << 17)) || cpu > 0)
#else
	/* if cpu is not cf, but library was compiled for it... */
	if (flags & (1L << 17))
#endif
	{
		slb_iconv_close();
		return -EINVAL;
	}
	
	ret = libiconv_slb_control(1, &iconvslb_funcs);
	if (ret < 0)
	{
		slb_iconv_close();
		return ret;
	}
	
	return ret;
}
